import pandas as pd
import numpy as np
import mdptoolbox

# Loading taxi data
file_path = 'formatted_yellow_taxi_22.parquet'
taxi_data = pd.read_parquet(file_path)

# State transition matrix part:

# Extracting all unique location IDs from the dataframe 
unique_locations = pd.unique(taxi_data[['PULocationID', 'DOLocationID']].values.ravel('K'))

# Creates the dictionary with all unique location IDs from the dataframe 
location_to_index = {loc_id: index for index, loc_id in enumerate(unique_locations)}

# Set up transition matrix with 2 dimensions 
num_locations = len(unique_locations)

# initializes the transition matrix with dimensions num_locations
transition_matrix = np.zeros((num_locations, num_locations))

# Group by PULocationID and DOLocationID and count transitions
transition_counts = taxi_data.groupby(['PULocationID', 'DOLocationID']).size().reset_index(name='count')

# Populate the transition matrix
for _, row in transition_counts.iterrows():
    pu_index = location_to_index[row['PULocationID']]
    do_index = location_to_index[row['DOLocationID']]
    transition_matrix[pu_index, do_index] = row['count']

# Normalize the matrix to get probabilities
transition_matrix = np.divide(transition_matrix, transition_matrix.sum(axis=1, keepdims=True), 
                              out=np.zeros_like(transition_matrix), 
                              where=transition_matrix.sum(axis=1, keepdims=True) != 0)

# Replace NaNs with a uniform distribution (or self-loop)
transition_matrix = np.nan_to_num(transition_matrix, nan=1.0 / num_locations)

# Validate the matrix and print if it is stochastic or not 
if np.allclose(transition_matrix.sum(axis=1), 1):
    print("The transition matrix is stochastic.")
else:
    print("The transition matrix is not stochastic.")

# Initialize 3D transition matrix
num_locations = len(unique_locations)
transition_matrix_3d = np.zeros((num_locations, num_locations, num_locations))

# Populate the 3D transition matrix
for i in range(num_locations):
    transition_matrix_3d[i, :, :] = transition_matrix

# Reward matrix part:

def calculate_rewards_optimized(taxi_data):
    # Setting constants for airport locations and fee
    JFK_AIRPORT_ID = 132
    LAGUARDIA_AIRPORT_ID = 138
    AIRPORT_FEE = 1.25

    # Get all unique location IDs from both PULocationID and DOLocationID
    all_locations = np.union1d(taxi_data['PULocationID'].unique(), taxi_data['DOLocationID'].unique())
    all_locations.sort()  # Sort for consistent indexing

    # Map location IDs to indices in the matrix
    location_to_index = {loc: idx for idx, loc in enumerate(all_locations)}

    # Initialize the reward matrix
    num_states = len(all_locations)
    reward_matrix = np.zeros((num_states, num_states))

    # Calculate net rewards
    taxi_data['net_reward'] = taxi_data['total_amount'] - taxi_data['extra'] - taxi_data['mta_tax'] - taxi_data['congestion_surcharge']

    # Adjust for airport fee
    taxi_data.loc[taxi_data['DOLocationID'].isin([JFK_AIRPORT_ID, LAGUARDIA_AIRPORT_ID]), 'net_reward'] -= AIRPORT_FEE

    # Adjust reward by trip distance
    taxi_data['adjusted_reward'] = taxi_data['net_reward'] * taxi_data['trip_distance']

    # Aggregate rewards for each state-action pair
    aggregated_rewards = taxi_data.groupby(['PULocationID', 'DOLocationID'])['adjusted_reward'].sum().reset_index()

    # Populate the reward matrix
    for _, row in aggregated_rewards.iterrows():
        pu_index = location_to_index[row['PULocationID']]
        do_index = location_to_index[row['DOLocationID']]
        reward_matrix[pu_index, do_index] = row['adjusted_reward']

    # Normalize the reward matrix to avoid extremely large values
    reward_matrix /= taxi_data.shape[0]

    return reward_matrix

reward_matrix = calculate_rewards_optimized(taxi_data)


# Setting up and solving the MDP with discount factor 0.9 by using mdptoolbox
discount_factor = 0.9
mdp_solver = mdptoolbox.mdp.PolicyIteration(transition_matrix_3d, reward_matrix, discount_factor)
mdp_solver.run()

# Setting up and solving the MDP with discount factor 0.5 by using mdptoolbox
discount_factor = 0.5
mdp_solver2 = mdptoolbox.mdp.PolicyIteration(transition_matrix_3d, reward_matrix, discount_factor)
mdp_solver2.run()

# Setting up and solving the MDP with discount factor 0.1 by using mdptoolbox
discount_factor = 0.1
mdp_solver3 = mdptoolbox.mdp.PolicyIteration(transition_matrix_3d, reward_matrix, discount_factor)
mdp_solver3.run()

# Extracting and displaing optimal policy with discount factor 0.9
optimal_policy = mdp_solver.policy
print("Optimal Policy discount factor 0.9:", optimal_policy)

# Extracting and displaing optimal policy with discount factor 0.5
optimal_policy2 = mdp_solver2.policy
print("Optimal Policy discount factor 0.5:", optimal_policy2)

# Extracting and displaing optimal policy with discount factor 0.1
optimal_policy3 = mdp_solver3.policy
print("Optimal Policy discount factor 0.1:", optimal_policy3)

# Check if policies are equal
policies_equal_0_9_0_5 = all(policy1 == policy2 for policy1, policy2 in zip(mdp_solver.policy, mdp_solver2.policy))
policies_equal_0_9_0_1 = all(policy1 == policy2 for policy1, policy2 in zip(mdp_solver.policy, mdp_solver3.policy))
policies_equal_0_5_0_1 = all(policy1 == policy2 for policy1, policy2 in zip(mdp_solver2.policy, mdp_solver3.policy))

# Print the comparison of results (the optimal policies)
if policies_equal_0_9_0_5:
    print("Optimal policies for discount factors 0.9 and 0.5 are equal.")
else:
    print("Optimal policies for discount factors 0.9 and 0.5 are not equal.")

if policies_equal_0_9_0_1:
    print("Optimal policies for discount factors 0.9 and 0.1 are equal.")
else:
    print("Optimal policies for discount factors 0.9 and 0.1 are not equal.")

if policies_equal_0_5_0_1:
    print("Optimal policies for discount factors 0.5 and 0.1 are equal.")
else:
    print("Optimal policies for discount factors 0.5 and 0.1 are not equal.")