#import libraries
import pandas as pd
import numpy as np
import mdptoolbox

# Define function mdp solver by 4 arguments
def run_mdp_solver(solver, transition_matrix, reward_matrix, discount_factor):
    mdp_solver = solver(transition_matrix, reward_matrix, discount_factor)
    mdp_solver.run()
    return mdp_solver.policy

# Function to save  the policies 
def save_policy(policy, filename):
    np.save(filename, policy)
    print(f"Saved policy to {filename}.npy")

# Function to print the policies 
def print_policy(policy, description):
    print(f"{description}: {policy}")

# Function to compare the policies 
def compare_policies(*policies):
    comparison_results = {}
    for i in range(len(policies)):
        for j in range(i + 1, len(policies)):
            comparison_results[(i, j)] = all(p1 == p2 for p1, p2 in zip(policies[i], policies[j]))
    return comparison_results

# Function to print comparison of the policies 
def print_comparison_results(comparison_results):
    for (i, j), result in comparison_results.items():
        if result:
            print(f"Optimal policies for instances {i} and {j} are equal.")
        else:
            print(f"Optimal policies for instances {i} and {j} are not equal.")

# Loading taxi data
file_path = 'yellow/formatted_yellow_taxi_22.parquet'
taxi_data = pd.read_parquet(file_path)

# State transition matrix part:

# Extracting all unique location IDs from the dataframe 
unique_locations = pd.unique(taxi_data[['PULocationID', 'DOLocationID']].values.ravel('K'))

# Creates the dictionary with all unique location IDs from the dataframe 
location_to_index = {loc_id: index for index, loc_id in enumerate(unique_locations)}

# Set up transition matrix with 2 dimensions 
num_locations = len(unique_locations)

# initializes the transition matrix with dimensions num_locations
transition_matrix = np.zeros((num_locations, num_locations))

# Group by PULocationID and DOLocationID and count transitions
transition_counts = taxi_data.groupby(['PULocationID', 'DOLocationID']).size().reset_index(name='count')

# Populate the transition matrix
for _, row in transition_counts.iterrows():
    pu_index = location_to_index[row['PULocationID']]
    do_index = location_to_index[row['DOLocationID']]
    transition_matrix[pu_index, do_index] = row['count']

# Normalize the matrix to get probabilities
transition_matrix = np.divide(transition_matrix, transition_matrix.sum(axis=1, keepdims=True), 
                              out=np.zeros_like(transition_matrix), 
                              where=transition_matrix.sum(axis=1, keepdims=True) != 0)

# Replace NaNs with a uniform distribution (or self-loop)
transition_matrix = np.nan_to_num(transition_matrix, nan=1.0 / num_locations)

# Validate the matrix and print if it is stochastic or not 
if np.allclose(transition_matrix.sum(axis=1), 1):
    print("The transition matrix is stochastic.")
else:
    print("The transition matrix is not stochastic.")

# Initialize 3D transition matrix
num_locations = len(unique_locations)
transition_matrix_3d = np.zeros((num_locations, num_locations, num_locations))

# Populate the 3D transition matrix
for i in range(num_locations):
    transition_matrix_3d[i, :, :] = transition_matrix

# Reward matrix part:

def calculate_rewards_optimized(taxi_data):
    # Setting constants for airport locations and fee
    JFK_AIRPORT_ID = 132
    LAGUARDIA_AIRPORT_ID = 138
    AIRPORT_FEE = 1.25

    # Get all unique location IDs from both PULocationID and DOLocationID
    all_locations = np.union1d(taxi_data['PULocationID'].unique(), taxi_data['DOLocationID'].unique())
    all_locations.sort()  # Sort for consistent indexing

    # Map location IDs to indices in the matrix
    location_to_index = {loc: idx for idx, loc in enumerate(all_locations)}

    # Initialize the reward matrix
    num_states = len(all_locations)
    reward_matrix = np.zeros((num_states, num_states))

    # Calculate net rewards
    taxi_data['net_reward'] = taxi_data['total_amount'] - taxi_data['extra'] - taxi_data['mta_tax'] - taxi_data['congestion_surcharge']

    # Adjust for airport fee
    taxi_data.loc[taxi_data['DOLocationID'].isin([JFK_AIRPORT_ID, LAGUARDIA_AIRPORT_ID]), 'net_reward'] -= AIRPORT_FEE

    # Adjust reward by trip distance
    taxi_data['adjusted_reward'] = taxi_data['net_reward'] * taxi_data['trip_distance']

    # Aggregate rewards for each state-action pair
    aggregated_rewards = taxi_data.groupby(['PULocationID', 'DOLocationID'])['adjusted_reward'].sum().reset_index()

    # Populate the reward matrix
    for _, row in aggregated_rewards.iterrows():
        pu_index = location_to_index[row['PULocationID']]
        do_index = location_to_index[row['DOLocationID']]
        reward_matrix[pu_index, do_index] = row['adjusted_reward']

    # Normalize the reward matrix to avoid extremely large values
    reward_matrix /= taxi_data.shape[0]

    return reward_matrix

reward_matrix = calculate_rewards_optimized(taxi_data)


# Set up discount factors
discount_factors = [0.9, 0.5, 0.1]
# Run Value iteration with different discount factors
policies_vi = []
for i, df in enumerate(discount_factors):
    policy = run_mdp_solver(mdptoolbox.mdp.ValueIteration, transition_matrix_3d, reward_matrix, df)
    policies_vi.append(policy)
    save_policy(policy, f"optimal_policy_vi_df_{df}")
    print_policy(policy, f"Optimal policy for Value iteration with discount factor {df}")

# Run Policy iteration with different discount factors
policies_pi = []
for i, df in enumerate(discount_factors):
    policy = run_mdp_solver(mdptoolbox.mdp.PolicyIteration, transition_matrix_3d, reward_matrix, df)
    policies_pi.append(policy)
    save_policy(policy, f"optimal_policy_pi_df_{df}")
    print_policy(policy, f"Optimal policy for Policy iteration with discount factor {df}")

# Run Q-Learning with different discount factors
policies_ql = []
for i, df in enumerate(discount_factors):
    policy = run_mdp_solver(mdptoolbox.mdp.QLearning, transition_matrix_3d, reward_matrix, df)
    policies_ql.append(policy)
    save_policy(policy, f"optimal_policy_ql_df_{df}")
    print_policy(policy, f"Optimal policy for Q-Learning with discount factor {df}")

# Compare all policies
comparison_results = compare_policies(*policies_vi, *policies_pi, *policies_ql)
print_comparison_results(comparison_results)