import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def load_taxi_data(file_path):
    try:
        return pd.read_parquet(file_path)
    except Exception as e:
        print(f"Error loading the file: {e}")
        return None

# Transition Matrix 
def initialize_matrices(taxi_data): 
    unique_locations = pd.unique(taxi_data[['PULocationID', 'DOLocationID']].values.ravel('K'))
    location_to_index = {loc_id: index for index, loc_id in enumerate(unique_locations)}
    num_locations = len(unique_locations)
    transition_matrix = np.zeros((num_locations, num_locations))
    transition_counts = taxi_data.groupby(['PULocationID', 'DOLocationID']).size().reset_index(name='count')

    for _, row in transition_counts.iterrows():
        pu_index = location_to_index[row['PULocationID']]
        do_index = location_to_index[row['DOLocationID']]
        transition_matrix[pu_index, do_index] = row['count']

    transition_matrix = np.divide(transition_matrix, transition_matrix.sum(axis=1, keepdims=True),
                                  out=np.zeros_like(transition_matrix),
                                  where=transition_matrix.sum(axis=1, keepdims=True) != 0)
    transition_matrix = np.nan_to_num(transition_matrix, nan=1.0 / num_locations)

    if np.allclose(transition_matrix.sum(axis=1), 1):
        print("The transition matrix is stochastic.")
    else:
        print("The transition matrix is not stochastic.")

# Reward Matrix Code
    reward_matrix = calculate_rewards_optimized(taxi_data)

    return transition_matrix, reward_matrix

def calculate_rewards_optimized(taxi_data):
    JFK_AIRPORT_ID = 132
    LAGUARDIA_AIRPORT_ID = 138
    AIRPORT_FEE = 1.25

    all_locations = np.union1d(taxi_data['PULocationID'].unique(), taxi_data['DOLocationID'].unique())
    all_locations.sort()

    location_to_index = {loc: idx for idx, loc in enumerate(all_locations)}

    num_states = len(all_locations)
    reward_matrix = np.zeros((num_states, num_states))

    taxi_data['net_reward'] = taxi_data['total_amount'] - taxi_data['extra'] - taxi_data['mta_tax'] - taxi_data['congestion_surcharge']
    taxi_data.loc[taxi_data['DOLocationID'].isin([JFK_AIRPORT_ID, LAGUARDIA_AIRPORT_ID]), 'net_reward'] -= AIRPORT_FEE
    taxi_data['adjusted_reward'] = taxi_data['net_reward'] * taxi_data['trip_distance']

    aggregated_rewards = taxi_data.groupby(['PULocationID', 'DOLocationID'])['adjusted_reward'].sum().reset_index()

    for _, row in aggregated_rewards.iterrows():
        pu_index = location_to_index[row['PULocationID']]
        do_index = location_to_index[row['DOLocationID']]
        reward_matrix[pu_index, do_index] = row['adjusted_reward']

    reward_matrix /= taxi_data.shape[0]

    return reward_matrix

# Setup  Q-table with zeros it maps state-action pairs to values
def sarsa_algorithm(transition_matrix, reward_matrix, num_states, num_actions, episodes, alpha, gamma):
    Q = np.zeros((num_states, num_actions))

    for episode in range(episodes):
        # Initialize state randomly for each new episode
        state = np.random.randint(0, num_states)

        # Choose action from state using policy from Q
        action = np.random.randint(0, num_actions)  # Example: random action

        for _ in range(1000):  # Limit number of steps per episode till 1000
            # Take action then observe new state and reward
            new_state = np.argmax(transition_matrix[state, action])  
            reward = reward_matrix[state, action]
                        
            # Choose new action from new state by using policy derived from Q
            new_action = np.random.randint(0, num_actions)  # Example: random action

            # Updating Q-value for  current state and action pair using  SARSA update rule - pair (new_state, new_action).
            Q[state, action] = Q[state, action] + alpha * (reward + gamma * Q[new_state, new_action] - Q[state, action])
            
            # Moving to new state and action for  next iteration
            state, action = new_state, new_action

    return Q

# Extract  optimal policy from  Q-table
def extract_policy(Q):
    #for each state choose action with highest Q-value
    return np.argmax(Q, axis=1)

# Plot the extracted policy as a function of states.
def plot_policy(policy, alpha, gamma):
    plt.figure(figsize=(12, 6))
    plt.plot(policy, marker='o')
    plt.title(f'Policy for alpha={alpha}, gamma={gamma}')
    plt.xlabel('States')
    plt.ylabel('Chosen Action')
    plt.grid(True)
    plt.show()

def main():
    file_path = 'yellow/formatted_yellow_taxi_22.parquet'
    taxi_data = load_taxi_data(file_path)

    if taxi_data is not None:
        # Plot the extracted policy as a function of states.
        transition_matrix, reward_matrix = initialize_matrices(taxi_data)
        
        alphas = [0.1, 0.5, 0.9] # Set up different learning rates
        gammas = [0.1, 0.5, 0.9] # Set up different discount factors
        episodes = 1000

        # Run SARSA algorithm with actual alpha and gamma values
        for alpha in alphas:
            for gamma in gammas:
                Q = sarsa_algorithm(transition_matrix, reward_matrix, len(transition_matrix), transition_matrix.shape[1], episodes, alpha, gamma)
                policy = extract_policy(Q)
                plot_policy(policy, alpha, gamma)

                # Save  policies to the main path
                policy_file_name = f"policy_alpha_{alpha}_gamma_{gamma}.npy"
                np.save(policy_file_name, policy)
                print(f"Policy saved as {policy_file_name}")
                                # Save the policy as a CSV file
                policy_df = pd.DataFrame(policy, columns=['Action'])
                csv_file_name = f"policy_alpha_{alpha}_gamma_{gamma}.csv"
                policy_df.to_csv(csv_file_name, index_label='State')
                print(f"Policy saved as {csv_file_name}")

if __name__ == "__main__":
    main()
