# Load libraries
import pandas as pd
import numpy as np

# Load the Parquet file
taxi_data = pd.read_parquet('yellow/formatted_yellow_taxi_22.parquet')

# Extracting all unique location IDs from the dataframe 
unique_locations = pd.unique(taxi_data[['PULocationID', 'DOLocationID']].values.ravel('K'))

# Creates the dictionary with all unique location IDs from the dataframe 
location_to_index = {loc_id: index for index, loc_id in enumerate(unique_locations)}

# Set up transition matrix with 2 dimensions 
num_locations = len(unique_locations)

# initializes the transition matrix with dimensions num_locations
transition_matrix = np.zeros((num_locations, num_locations))

# Group by PULocationID and DOLocationID and count transitions
transition_counts = taxi_data.groupby(['PULocationID', 'DOLocationID']).size().reset_index(name='count')

# Populate the transition matrix
for _, row in transition_counts.iterrows():
    pu_index = location_to_index[row['PULocationID']] #retrives each Pickup and dropoff location ids indecies 
    do_index = location_to_index[row['DOLocationID']]
    transition_matrix[pu_index, do_index] = row['count']

# Normalize the matrix to get probabilities by converting counts into probabilities of transitioning from one location to another and then insures that each row sum equals 1
transition_matrix = np.divide(transition_matrix, transition_matrix.sum(axis=1, keepdims=True), 
                              out=np.zeros_like(transition_matrix), 
                              where=transition_matrix.sum(axis=1, keepdims=True) != 0)

# Replace nans with uniform distribution.
transition_matrix = np.nan_to_num(transition_matrix, nan=1.0 / num_locations)

# Validate the matrix and print if it is stochastic or not 
if np.allclose(transition_matrix.sum(axis=1), 1):
    print("The transition matrix is stochastic.")
else:
    print("The transition matrix is not stochastic.")
